//using System.Net;
using System;
using System.Net.Sockets;

namespace rde
{
    public class Common
    {
        public enum ServerToClientMsg
        {
            TOOL_EXECUTABLE,
            FILE_NAME,
            START_PROCESS,
            WORKING_MODE,
            FILE_LIST_COMPLETE,
            GROUP_DESCRIPTION,
            TASK_DESCRIPTION,
            BREAK
        };
        public enum ClientToServerMsg
        {
            MACHINE_NAME,
            FILE_DONE,
            ASK_FOR_WORK,
            CLIENT_STATE,
            TOOL_ERROR,
            TASK_DONE,
            CLIENT_CLOSING
        };
        public enum ClientState
        {
            IDLE,
            WORKING
        };
        public enum ClientWorkingMode
        {
            LOAD_BALANCING,
            BATCH
        };
        public const int kDefaultPort = 1110;

        public static void PutShort(byte[] msg, int index, short value)
        {
            msg[index] = (byte)((value >> 8) & 0xFF);
            msg[index + 1] = (byte)(value & 0xFF); 
        }
        public static int PutString(byte[] msg, string s, int i)
        {
            short len = (s == null ? (short)0 : (short)s.Length);
            rde.Common.PutShort(msg, i, len);
            if (len > 0)
                System.Text.Encoding.ASCII.GetBytes(s, 0, len, msg, i + 2);
            return len + 2;
        }

        public static string GetASCIIZStringFromData(byte[] msg, int index)
        {
            System.Text.Decoder d = System.Text.Encoding.ASCII.GetDecoder();
            char[] chars = new char[msg.Length];
            d.GetChars(msg, index, msg.Length - index, chars, 0);
            string ret = new string(chars);
            char[] zeroChars = { '\0' };
            return ret.TrimEnd(zeroChars);
        }
        public static string GetASCIIZStringFromData(byte[] msg)
        {
            return GetASCIIZStringFromData(msg, 1);
        }


        public static bool SendMessage(Socket socket, byte[] msg, int msgSize)
        {
            System.Diagnostics.Debug.Assert(msgSize < 65536);
            System.Diagnostics.Debug.Assert(msgSize <= msg.Length);
            byte[] markedMessage = new byte[2 + msgSize];
            markedMessage[0] = (byte)((msgSize >> 8) & 0xFF);
            markedMessage[1] = (byte)(msgSize & 0xFF);
            Array.Copy(msg, 0, markedMessage, 2, msgSize);
            try
            {
                socket.Send(markedMessage);
            }
            catch (SocketException)
            {
                if (socket != null && socket.Connected)
                    socket.Close();
                return false;
            }
            return true;
        }
        public static bool SendMessage(Socket socket, byte[] msg)
        {
            return SendMessage(socket, msg, msg.Length);
        }
    }
}
