﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Xml;
using System.Xml.Serialization;
using System.Threading;

namespace rdeserver
{
    public partial class rdeserver : Form
    {
        const int kMaxLastTaskResults = 24;
        
        public delegate void DelegateAddClient(Socket socket);
        public delegate void DelegateSetClientState(int client, rde.Common.ClientState state);
        public delegate void DelegateSetClientName(int client, string name);
        public delegate void DelegateUpdateClientList();
        public delegate void DelegateDebugMessage(string msg);
        public delegate void DelegateSetStatus(string msg);

        class ClientDesc
        {
            public int id;
            public string name;
            public string label;
            public rde.Common.ClientState state = rde.Common.ClientState.IDLE;
            public Socket socket;
        };

        public rdeserver()
        {
            InitializeComponent();
            InitTasks();
        }

        void InitTasks()
        {
            if (tasksToolStripMenuItem.DropDownItems.Count > 0)
            {
                ToolStripItem breakTaskItem = tasksToolStripMenuItem.DropDownItems[0];
                tasksToolStripMenuItem.DropDownItems.Clear();
                tasksToolStripMenuItem.DropDownItems.Add(breakTaskItem);
            }

            LoadTasks("test.xml");
            foreach (Task t in m_tasks)
            {
                ToolStripMenuItem taskItem = new ToolStripMenuItem();
                taskItem.Text = t.name;
                taskItem.Tag = t;
                taskItem.Enabled = false;
                taskItem.Click += new EventHandler(taskItem_Click);
                taskItem.ToolTipText = t.description;
                tasksToolStripMenuItem.DropDownItems.Add(taskItem);
            }
        }

        void EnableTasks(bool enable, bool enableAll)
        {
            //foreach (ToolStripMenuItem it in tasksToolStripMenuItem.DropDownItems)
            for (int i = 0; i < tasksToolStripMenuItem.DropDownItems.Count; ++i)
            {
                ToolStripItem it = tasksToolStripMenuItem.DropDownItems[i];
                if (enableAll || i > 0)
                    it.Enabled = enable;
            }
        }

        void taskItem_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem taskItem = sender as ToolStripMenuItem;
            Task t = taskItem.Tag as Task;
            ExecuteTask(t);
        }

        private void butStartListening_Click(object sender, EventArgs e)
        {
            try
            {
                //Task t = new Task();
                //t.name = "nvcompress";

                //Task.EntryGroup grp = new Task.EntryGroup();
                //t.workingDir = "c:\\!projects\\textures";
                //grp.files = new List<string>();
                //grp.files.Add("lena_std.tga");
                //t.AddGroup(grp);
                //grp = new Task.EntryGroup();
                //grp.extensions = new List<string>();
                //grp.extensions.Add(".tga");
                //grp.extensions.Add(".jpg");
                //grp.commandLine = "$(SourceFile)";
                //grp.SendDescription(m_serverSocket);
                //t.AddGroup(grp);

                /*t.sourceFiles.Add(new Task.SourceFile("lena_std.tga"));
                t.sourceExtensions.Add(new Task.Extension(".tga"));
                t.sourceExtensions.Add(new Task.Extension(".jpg"));*/
                //t.PrepareForExecution();
                ////t.ProcessFiles();
                //m_tasks.Add(t);
                //SaveTasks("test.xml");
                //t.PrepareForExecution();
                //filesToProcess = t.GetFilesToProcess();

                //foreach (Task tsk in m_tasks)
                //{
                //    //cbCommands.Items.Add(t.name);
                //    ToolStripMenuItem taskItem = new ToolStripMenuItem();
                //    taskItem.Text = tsk.name;
                //    taskItem.Tag = tsk;
                //    taskItem.Enabled = false;
                //    taskItem.Click += new EventHandler(taskItem_Click);
                //    taskItem.ToolTipText = tsk.description;
                //    tasksToolStripMenuItem.DropDownItems.Add(taskItem);

                //}

                if (textBoxPort.Text == "")
                {
                    MessageBox.Show("Invalid listening port");
                    return;
                }

                m_serverSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                int port = System.Convert.ToInt32(textBoxPort.Text);
                IPEndPoint ip = new IPEndPoint(IPAddress.Any, port);
                m_serverSocket.Bind(ip);
                m_serverSocket.Listen(16);
                m_serverSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
                butStartListening.Enabled = false;
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        public void OnClientConnect(IAsyncResult res)
        {
            try
            {
                Socket clientSocket = m_serverSocket.EndAccept(res);
                Invoke(new DelegateAddClient(AddClient), new Object[] { clientSocket });

                WaitForData(clientSocket, m_clients.Count - 1);
                m_serverSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        void SaveTask(Task t, string fileName)
        {
            XmlSerializer s = new XmlSerializer(t.GetType());
            TextWriter w = new StreamWriter(fileName, false);
            s.Serialize(w, t);
            w.Close();
        }

        void LoadTasks(string fileName)
        {
            XmlSerializer s = new XmlSerializer(m_tasks.GetType());
            TextReader r = new StreamReader(fileName);
            m_tasks = (List<Task>)s.Deserialize(r);
            r.Close();
            foreach (Task t in m_tasks)
                t.PostProcess();
        }
        void SaveTasks(string fileName)
        {
            XmlSerializer s = new XmlSerializer(m_tasks.GetType());
            TextWriter w = new StreamWriter(fileName, false);
            s.Serialize(w, m_tasks);
            w.Close();
        }
        void AddClient(Socket socket)
        {
            ClientDesc desc = new ClientDesc();
            int numClients = m_clients.Count;
            desc.id = numClients;
            desc.label = "Client: " + numClients.ToString();
            desc.socket = socket;
            lbClients.Items.Add(desc.label);
            m_clients.Add(desc);
        }

        class SocketPacket
        {
            public SocketPacket(Socket socket, int id)
            {
                clientSocket = socket;
                clientId = id;
            }
            public Socket clientSocket;
            public int clientId;
            public byte[] dataBuffer = new byte[2048];
        };

        void WaitForData(Socket socket, int clientId)
        {
            try
            {
                if (m_dataCallback == null)
                    m_dataCallback = new AsyncCallback(OnDataReceived);

                SocketPacket packet = new SocketPacket(socket, clientId);
                socket.BeginReceive(packet.dataBuffer, 0, packet.dataBuffer.Length,
                    SocketFlags.None, m_dataCallback, packet);
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        void SendSimpleMessage(Socket socket, rde.Common.ServerToClientMsg msgId)
        {
            byte[] data = new byte[1];
            data[0] = (byte)msgId;
            rde.Common.SendMessage(socket, data);
        }

        void SendFileToProcess(Socket socket, string fileName)
        {
            byte[] buffer = new byte[fileName.Length + 10];
            buffer[0] = (byte)rde.Common.ServerToClientMsg.FILE_NAME;
            System.Text.Encoding.ASCII.GetBytes(fileName, 0, fileName.Length,
                buffer, 1);
            rde.Common.SendMessage(socket, buffer, buffer.Length);
        }

        void HandleMessage(SocketPacket packet, byte[] msg)
        {
            rde.Common.ClientToServerMsg msgType = (rde.Common.ClientToServerMsg)msg[0];
            if (msgType == rde.Common.ClientToServerMsg.MACHINE_NAME)
            {
                string clientName = rde.Common.GetASCIIZStringFromData(msg);

                Invoke(new DelegateSetClientName(SetClientName),
                    new Object[] { packet.clientId, clientName });
            }
            else if (msgType == rde.Common.ClientToServerMsg.CLIENT_STATE)
            {
                rde.Common.ClientState state = (rde.Common.ClientState)msg[1];
                Invoke(new DelegateSetClientState(SetClientState),
                    new Object[] { packet.clientId, state });
            }
            else if (msgType == rde.Common.ClientToServerMsg.ASK_FOR_WORK)
            {
                if (m_filesToProcess.Count != 0)
                {
                    Task.Entry file = m_filesToProcess.Dequeue();
                    //Console.WriteLine("Sending {0} to {1}", file.fileName, packet.clientId);
                    file.Send(packet.clientSocket);
                }
                else
                {
                    // Work done.
                    SendSimpleMessage(packet.clientSocket, rde.Common.ServerToClientMsg.FILE_LIST_COMPLETE);
                }
            }
            else if (msgType == rde.Common.ClientToServerMsg.CLIENT_CLOSING)
            {
                packet.clientSocket.Close();
                Invoke(new DelegateUpdateClientList(UpdateClientList), null);
            }
            else if (msgType == rde.Common.ClientToServerMsg.TOOL_ERROR)
            {
                string errMsg = rde.Common.GetASCIIZStringFromData(msg);
                m_lastTaskResults[m_lastTaskResults.Count - 1].errorMessages.Add(errMsg);
                Invoke(new DelegateDebugMessage(AddDebugMessage), new Object[] { errMsg });

                if (m_activeTask != null && !m_activeTask.ignoreErrors)
                {
                    m_lastTaskResults[m_lastTaskResults.Count - 1].code = TaskResult.Code.FAILED;
                    BreakActiveTask();
                }
            }
            else if (msgType == rde.Common.ClientToServerMsg.TASK_DONE)
            {
                if (m_filesToProcess.Count == 0)
                {
                    TaskResult res = m_lastTaskResults[m_lastTaskResults.Count - 1];
                    if (res.code == TaskResult.Code.IN_PROGRESS)
                        res.code = TaskResult.Code.COMPLETED;
                    res.endTime = DateTime.Now;
                    m_activeTask = null;
                    Invoke(new DelegateSetStatus(SetStatus), new Object[] { "" });
                }
            }
            else
            {
                System.Diagnostics.Debug.Fail("Unknown command ID: " + msgType);
            }
        }

        void OnDataReceived(IAsyncResult res)
        {
            SocketPacket packet = (SocketPacket)res.AsyncState;
            try
            {
                int receivedBytes = packet.clientSocket.EndReceive(res);

                List<byte[]> messages =
                    m_msgReceiver.OnDataReceived(packet.dataBuffer, receivedBytes);
                for (int i = 0; i < messages.Count; ++i)
                {
                    HandleMessage(packet, messages[i]);
                }

                if (packet.clientSocket != null && packet.clientSocket.Connected)
                    WaitForData(packet.clientSocket, packet.clientId);
            }
            catch (ObjectDisposedException)
            {
                Invoke(new DelegateUpdateClientList(UpdateClientList), null);
            }
            catch (SocketException exc)
            {
                MessageBox.Show(exc.Message);
                Invoke(new DelegateUpdateClientList(UpdateClientList), null);
            }
        }

        void UpdateClientLabel(int client)
        {
            string[] stateDesc = { "IDLE", "WORKING" };

            ClientDesc desc = m_clients[client];
            desc.label = client.ToString() + ": " + desc.name + " - " +
                stateDesc[(int)desc.state];
            lbClients.Items[client] = desc.label;
            UpdateClientList();
        }

        void SetClientState(int client, rde.Common.ClientState state)
        {
            ClientDesc desc = m_clients[client];
            desc.state = state;
            UpdateClientLabel(client);
        }
        void SetClientName(int client, string name)
        {
            ClientDesc desc = m_clients[client];
            desc.name = name;
            UpdateClientLabel(client);
        }
        void UpdateClientList()
        {
            lbClients.Items.Clear();
            for (int i = 0; i < m_clients.Count; ++i)
            {
                if (m_clients[i].socket != null && m_clients[i].socket.Connected)
                    lbClients.Items.Add(m_clients[i].label);
            }
            RemoveInactiveClients();
        }

        public class TaskResult
        {
            public enum Code
            {
                COMPLETED,
                FAILED,
                INTERRUPTED,
                IN_PROGRESS
            };

            public Code code;
            public string taskName;
            public int numItems;
            public DateTime startTime;
            public DateTime endTime;
            public int numClients;
            public int numClientsDone = 0;
            public List<String> errorMessages = new List<String>();
        };

        Socket m_serverSocket;
        List<Task> m_tasks = new List<Task>();
        Task m_activeTask = null;
        List<ClientDesc> m_clients = new List<ClientDesc>();
        AsyncCallback m_dataCallback;
        rde.MsgReceiver m_msgReceiver = new rde.MsgReceiver();
        Queue<Task.Entry> m_filesToProcess;
        List<TaskResult> m_lastTaskResults = new List<TaskResult>();

        private void textBoxPort_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsNumber(e.KeyChar))
                e.Handled = true;
        }

        private bool IsClientInactive(ClientDesc desc)
        {
            return desc.socket == null || !desc.socket.Connected;
        }
        void RemoveInactiveClients()
        {
            m_clients.RemoveAll(IsClientInactive);
        }

        void SetClientWorkingMode(Socket socket, rde.Common.ClientWorkingMode mode)
        {
            byte[] msg = new byte[2];
            msg[0] = (byte)rde.Common.ServerToClientMsg.WORKING_MODE;
            msg[1] = (byte)mode;
            rde.Common.SendMessage(socket, msg);
        }

        void ExecuteTask(Task t)
        {
            t.PrepareForExecution();
            m_filesToProcess = t.GetFilesToProcess();
            RemoveInactiveClients();
            SetStatus("Executing task: " + t.name);
            for (int i = 0; i < m_clients.Count; ++i)
            {
                Socket s = m_clients[i].socket;
                
                t.SendDescription(s);
                t.SendGroupDescriptions(s);
                t.SendPreBuildStep(s);
                SetClientWorkingMode(s, t.batchProcessing ?
                    rde.Common.ClientWorkingMode.BATCH :
                    rde.Common.ClientWorkingMode.LOAD_BALANCING);

                // That's all... Client's will ask for files to process by 
                // themselves.
            }
            m_activeTask = t;
            TaskResult result = new TaskResult();
            result.taskName = t.name;
            result.code = TaskResult.Code.IN_PROGRESS;
            result.numItems = m_filesToProcess.Count;
            result.startTime = DateTime.Now;
            result.numClients = m_clients.Count;
            if (m_lastTaskResults.Count + 1 >= kMaxLastTaskResults)
                m_lastTaskResults.RemoveAt(0);
            m_lastTaskResults.Add(result);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            UpdateClientList();
            EnableTasks(m_clients.Count != 0 && m_activeTask == null, m_activeTask == null);
            //butStartListening.Enabled = (m_numClients == 0);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        void SetStatus(string status)
        {
            labelStatus.Text = status;
        }

        void AddDebugMessage(string msg)
        {
            textBoxConsole.AppendText(msg + "\r\n");
        }

        private void label1_Click(object sender, EventArgs e)
        {

        }

        void BreakActiveTask()
        {
            foreach (ClientDesc desc in m_clients)
                SendSimpleMessage(desc.socket, rde.Common.ServerToClientMsg.BREAK);
            m_filesToProcess.Clear();
        }

        private void breakTaskToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_activeTask != null)
            {
                m_lastTaskResults[m_lastTaskResults.Count - 1].code = TaskResult.Code.INTERRUPTED;
                BreakActiveTask();
                AddDebugMessage("Trying to interrupt " + m_activeTask.name);
            }
        }

        private void toolStripMenuItem1_Click(object sender, EventArgs e)
        {

        }

        private void lastTasksToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LastTasks lastTasksDlg = new LastTasks(m_lastTaskResults);
            lastTasksDlg.ShowDialog(this);
        }

        private void reloadTasksToolStripMenuItem_Click(object sender, EventArgs e)
        {
            InitTasks();
        }

    }
}
